/*
************************************************************************
*
*   MotOGLDump.c - Motif/OpenGL image dump
*
*   Copyright (c) 1994
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/05/09
*   Pathname of SCCS file     : /sgiext/molmol/src/motogl/SCCS/s.MotOGLDump.c
*   SCCS identification       : 1.10
*
************************************************************************
*/

#include "motogl.h"

#include <stdio.h>
#include <stdlib.h>
#define BOOL X11_BOOL  /* hack to avoid naming conflict */
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <GL/gl.h>
#include <GL/glx.h>
#undef BOOL

#include <math.h>
#include <tiffio.h>

#include "motogl_int.h"
#include <sg.h>
#include <sg_get.h>
#include <sg_attr.h>
#include <io_error.h>

/* X servers often grow bigger and bigger when allocating/freeing
   many pixmaps, so it's better to keep and reuse them if possible. */
#define KEEP_PIXMAP 1

static FILE *TiffFileP;
static SgPlotOrient Orient;
static int ImgW, ImgH;
static float Gamma;
static BOOL OutOfMemory;
static Display *Dpy;
static Pixmap XPix = 0;
static GLXPixmap GPix = 0;
static GLXContext OldCtx, Ctx;
static float OldVpX, OldVpY, OldVpW, OldVpH;

static void
destroyPixmap(void)
{
  glXDestroyGLXPixmap(Dpy, GPix);
  GPix = 0;
  XFreePixmap(Dpy, XPix);
  XPix = 0;
}

static int
xErrorHandler(Display *dpy, XErrorEvent *evtP)
{
  OutOfMemory = TRUE;

  return 0;
}

IORes
IOMotifOGLStartDump(char *fileName,
    SgPlotOrient orient, int w, int h, float gamma)
{
  Widget drawW = MotifOGLGetDrawW();
  XErrorHandler oldHandler;
  int attrList[20];
  int n;
  XVisualInfo *visP;
  GLfloat v[4];
  int i;

  TiffFileP = fopen(fileName, "w");
  if (TiffFileP == NULL) {
    IORaiseError("could not open output file");
    return IO_RES_ERR;
  }

#if KEEP_PIXMAP
  if (GPix != 0 && (w != ImgW || h != ImgH))
    destroyPixmap();
#endif

  Orient = orient;
  ImgW = w;
  ImgH = h;
  Gamma = gamma;

  Dpy = XtDisplay(drawW);

  n = 0;
  attrList[n++] = GLX_RGBA;
  attrList[n++] = GLX_RED_SIZE; attrList[n++] = 8;
  attrList[n++] = GLX_GREEN_SIZE; attrList[n++] = 8;
  attrList[n++] = GLX_BLUE_SIZE; attrList[n++] = 8;
  attrList[n++] = GLX_DEPTH_SIZE; attrList[n++] = 1;
  attrList[n++] = None;
  visP = glXChooseVisual(Dpy,
      XScreenNumberOfScreen(XtScreen(drawW)), attrList);
  if (visP == NULL) {
    IORaiseError("no 24-bit true color visual available");
    return IO_RES_ERR;
  }

  /* catch BadAlloc error */
  OutOfMemory = FALSE;
  oldHandler = XSetErrorHandler(xErrorHandler);

  if (XPix == 0) {
    XPix = XCreatePixmap(Dpy, XtWindow(drawW), w, h, 24);
    XSync(Dpy, FALSE);  /* error comes too late otherwise */
    if (OutOfMemory) {
      XPix = 0;
      XSetErrorHandler(oldHandler);
      IORaiseError("could not allocate Pixmap");
      return IO_RES_ERR;
    }
  }

  if (GPix == 0) {
    GPix = glXCreateGLXPixmap(Dpy, visP, XPix);
    XSync(Dpy, FALSE);
    XSetErrorHandler(oldHandler);
    if (OutOfMemory) {
      GPix = 0;
      XFreePixmap(Dpy, XPix);
      XPix = 0;
      IORaiseError("could not allocate Pixmap");
      return IO_RES_ERR;
    }
  }

  Ctx = glXCreateContext(Dpy, visP, NULL, False);
  if (Ctx == NULL) {
    destroyPixmap();
    IORaiseError("could not create rendering context");
    return IO_RES_ERR;
  }

  XFree((void *) visP);

  OldCtx = glXGetCurrentContext();
  (void) glXMakeCurrent(Dpy, GPix, Ctx);

  SgSetAttr();
  SgUpdateView();

  SgGetViewport(&OldVpX, &OldVpY, &OldVpW, &OldVpH);
  SgSetViewport(0.0, 0.0, w, h);

  for (i = 0; i < SG_FEATURE_NO; i++)
    SgSetFeature(i, SgGetFeature(i));

  /* SgOGLInit() is not called for new context! */
  glEnable(GL_NORMALIZE);
  v[0] = 1.0;
  v[1] = 1.0;
  v[2] = 1.0;
  v[3] = 1.0;
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, v);

  (void) MotifOGLPrepareFont(Dpy);

  return IO_RES_OK;
}

static IORes
writeTiff(void)
{
  TIFF *tif;
  int tiffW, tiffH;
  int rowsPerStrip, bufSize, rowI, i;
  float *fBuf;
  unsigned char *cBuf;
  IORes res;

  tif = TIFFFdOpen(fileno(TiffFileP), "output file", "w");
  if (tif == NULL) {
    IORaiseError("could not create TIFF file");
    return IO_RES_ERR;
  }

  if (Orient == SG_PLOT_NORMAL) {
    tiffW = ImgW;
    tiffH = ImgH;
    bufSize = 4 * ((3 * tiffW + 3) / 4);
    glPixelStorei(GL_PACK_ALIGNMENT, 4);
  } else {
    tiffW = ImgH;
    tiffH = ImgW;
    bufSize = 3 * tiffW;
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
  }

  rowsPerStrip = (8 * 1024) / (3 * tiffW);
  if (rowsPerStrip == 0)
    rowsPerStrip = 1;

  TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, tiffW);
  TIFFSetField(tif, TIFFTAG_IMAGELENGTH, tiffH);
  TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
  TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(tif, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
  TIFFSetField(tif, TIFFTAG_DOCUMENTNAME, "MOLMOL Plot");
  TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "MOLMOL Plot");
  TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 3);
  TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);
  TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);

  cBuf = malloc(bufSize * sizeof(*cBuf));
  if (Gamma != 1.0)
    fBuf = malloc(bufSize * sizeof(*fBuf));

  res = IO_RES_OK;
  for (rowI = 0; rowI < tiffH; rowI++) {
    if (Gamma == 1.0) {
      if (Orient == SG_PLOT_NORMAL)
	glReadPixels(0, ImgH - 1 - rowI, ImgW, 1,
	    GL_RGB, GL_UNSIGNED_BYTE, cBuf);
      else
	glReadPixels(rowI, 0, 1, ImgH,
	    GL_RGB, GL_UNSIGNED_BYTE, cBuf);
    } else {
      if (Orient == SG_PLOT_NORMAL)
	glReadPixels(0, ImgH - 1 - rowI, ImgW, 1,
	    GL_RGB, GL_FLOAT, fBuf);
      else
	glReadPixels(rowI, 0, 1, ImgH,
	    GL_RGB, GL_FLOAT, fBuf);
      
      for (i = 0; i < 3 * tiffW; i++)
	cBuf[i] = (int) (255.99 * powf(fBuf[i], Gamma));
    }

    if (TIFFWriteScanline(tif, cBuf, rowI, 0) < 0) {
      IORaiseError("error while writing TIFF file");
      res = IO_RES_ERR;
      break;
    }
  }

  free(cBuf);
  if (Gamma != 1.0)
    free(fBuf);

  TIFFFlushData(tif);
  TIFFClose(tif);

  return res;
}

IORes
IOMotifOGLEndDump(void)
{
  IORes res;

  res = writeTiff();

  (void) fclose(TiffFileP);

  (void) glXMakeCurrent(Dpy, XtWindow(MotifOGLGetDrawW()), OldCtx);
  SgSetAttr();
  SgSetViewport(OldVpX, OldVpY, OldVpW, OldVpH);

#if KEEP_PIXMAP
#else
  destroyPixmap();
#endif

  glXDestroyContext(Dpy, Ctx);

  return res;
}
